# -*- coding: utf-8 -*-

from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.views import APIView

from rest_search.schemas import get_form_schema, get_form_schema_operation_parameters


class SearchFilterBackend(BaseFilterBackend):
    """
    Dummy filter backend to enable API documentation.
    """

    def get_schema_fields(self, view):
        return get_form_schema(view.form_class)

    def get_schema_operation_parameters(self, view):
        return get_form_schema_operation_parameters(view.form_class)


class SearchAPIView(APIView):
    filter_backends = (SearchFilterBackend,)

    def get(self, request, *args, **kwargs):
        query = self.get_query()
        sort = self.get_sort()

        pagination = LimitOffsetPagination()
        pagination.default_limit = 20
        pagination.limit = pagination.get_limit(request)
        pagination.offset = pagination.get_offset(request)
        pagination.request = request

        body = {"query": query, "size": pagination.limit, "from": pagination.offset}
        if sort:
            body["sort"] = sort

        # execute elasticsearch query
        indexer = self.get_indexer()
        res = indexer.search(body=body)

        # map back to expected format
        items = list(indexer.map_results(res["hits"]["hits"]))
        pagination.count = res["hits"]["total"]
        return pagination.get_paginated_response(items)

    def get_indexer(self):
        return self.indexer_class()

    def get_query(self):
        """
        Returns the 'query' element of the ElasticSearch request body.
        """
        form = self.form_class(self.request.GET, context={"request": self.request})
        if not form.is_valid():
            raise ValidationError(form.errors)
        return form.get_query()

    def get_sort(self):
        """
        Returns the 'sort' element of the ElastiSearch request body.

        To use the default sorting (by score) return None.
        """
        return None



# -*- coding: utf-8 -*-

import functools

from rest_search import queue_flush


def flush_updates(function):
    """
    Decorator that flushes ElasticSearch updates.
    """

    @functools.wraps(function)
    def wrap(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        finally:
            queue_flush()

    return wrap



# -*- coding: utf-8 -*-

import coreapi
import coreschema
from django import forms
from django.utils.encoding import force_text


def get_form_schema(form_class):
    """
    Return the coreapi schema for the given form class.
    """
    fields = []
    for name, field in form_class.declared_fields.items():
        fields.append(
            coreapi.Field(
                location="query",
                name=name,
                required=field.required,
                schema=get_form_field_coreapi_schema(field),
            )
        )
    return fields


def get_form_field_coreapi_schema(field):
    """
    Returns the coreapi schema for the given form field.
    """
    title = force_text(field.label) if field.label else ""
    description = force_text(field.help_text) if field.help_text else ""

    if isinstance(field, forms.BooleanField):
        field_class = coreschema.Boolean
    elif isinstance(field, forms.DateTimeField):
        return coreschema.String(
            description=description, title=title, format="date-time"
        )
    elif isinstance(field, forms.FloatField):
        field_class = coreschema.Number
    elif isinstance(field, (forms.IntegerField, forms.ModelChoiceField)):
        field_class = coreschema.Integer
    else:
        field_class = coreschema.String

    return field_class(description=description, title=title)


def get_form_schema_operation_parameters(form_class):
    """
    Return the openapi schema for the given form class.
    """
    fields = []
    for name, field in form_class.declared_fields.items():
        fields.append(
            {
                "name": name,
                "required": field.required,
                "in": "query",
                "description": force_text(field.help_text) if field.help_text else "",
                "schema": get_form_field_openapi_schema(field),
            }
        )
    return fields


def get_form_field_openapi_schema(field):
    if isinstance(field, forms.BooleanField):
        return {"type": "boolean"}
    elif isinstance(field, forms.DateField):
        return {"type": "string", "format": "date"}
    elif isinstance(field, forms.DateTimeField):
        return {"type": "string", "format": "date-time"}
    elif isinstance(field, forms.FloatField):
        return {"type": "number"}
    elif isinstance(field, (forms.IntegerField, forms.ModelChoiceField)):
        return {"type": "integer"}
    else:
        return {"type": "string"}



# -*- coding: utf-8 -*-

import logging

from celery import shared_task
from django.conf import settings
from elasticsearch.helpers import bulk

from rest_search import DEFAULT_INDEX_SETTINGS, get_elasticsearch
from rest_search.indexers import _get_registered

logger = logging.getLogger("rest_search")


def create_index():
    """
    Creates the ElasticSearch indices if they do not exist.
    """
    for indexer in _get_registered():
        _create_index(indexer)


def delete_index():
    """
    Deletes the ElasticSearch indices.
    """
    for indexer in _get_registered():
        es = get_elasticsearch(indexer)
        es.indices.delete(index=indexer.index, ignore=404)


@shared_task
def patch_index(updates):
    """
    Performs a partial update of the ElasticSearch indices.
    """
    updates_str = ["%s: %d items" % (k, len(v)) for k, v in updates.items()]
    logger.info("Patching indices (%s)" % ", ".join(updates_str))

    indexers = _get_registered()
    for doc_type, pks in updates.items():
        for indexer in indexers:
            if indexer.doc_type == doc_type:
                _patch_index(indexer, pks)


@shared_task
def update_index(remove=True):
    """
    Performs a full update of the ElasticSearch indices.
    """
    logger.info("Updating indices")

    create_index()

    for indexer in _get_registered():
        _update_index(indexer, remove=remove)


def _create_index(indexer):
    body = {
        "mappings": {},
        "settings": getattr(
            settings, "REST_SEARCH_INDEX_SETTINGS", DEFAULT_INDEX_SETTINGS
        ),
    }
    if indexer.mappings is not None:
        body["mappings"][indexer.doc_type] = indexer.mappings

    es = get_elasticsearch(indexer)
    if not es.indices.exists(indexer.index):
        logger.info("Creating index %s" % indexer.index)
        es.indices.create(index=indexer.index, body=body)


def _delete_items(indexer, pks):
    es = get_elasticsearch(indexer)

    def mapper(pk):
        return {
            "_index": indexer.index,
            "_type": indexer.doc_type,
            "_id": pk,
            "_op_type": "delete",
        }

    bulk(es, map(mapper, pks), raise_on_error=False)


def _index_items(indexer, pks):
    es = get_elasticsearch(indexer)
    seen_pks = set()

    def bulk_mapper(block_size=1000):
        for i in range(0, len(pks), block_size):
            chunk = pks[i : i + block_size]
            qs = indexer.get_queryset().filter(pk__in=chunk)
            data = indexer.serializer_class(qs, many=True).data
            for item in data:
                seen_pks.add(item["id"])
                yield {
                    "_index": indexer.index,
                    "_type": indexer.doc_type,
                    "_id": item["id"],
                    "_source": item,
                }

    bulk(es, bulk_mapper())
    return seen_pks


def _patch_index(indexer, pks):
    # index current items
    seen_pks = _index_items(indexer, pks)

    # remove obsolete items
    removed_pks = set(pks) - seen_pks
    _delete_items(indexer, removed_pks)


def _update_index(indexer, remove):
    scan = indexer.scan(query={"stored_fields": []})
    old_pks = set([int(i["_id"]) for i in scan])

    # index current items
    seen_pks = _index_items(
        indexer, sorted(indexer.get_queryset().values_list("pk", flat=True))
    )

    # remove obsolete items
    if remove:
        removed_pks = old_pks - seen_pks
        _delete_items(indexer, removed_pks)



# -*- coding: utf-8 -*-

from threading import local

from aws_requests_auth.aws_auth import AWSRequestsAuth
from django.conf import settings
from elasticsearch import Elasticsearch, RequestsHttpConnection

DEFAULT_INDEX_SETTINGS = {
    "analysis": {
        "analyzer": {
            "default": {
                "tokenizer": "standard",
                "filter": ["standard", "lowercase", "asciifolding"],
            }
        }
    }
}
QUEUES = {}


class ConnectionHandler(object):
    def __init__(self):
        self._connections = local()

    def __delitem__(self, alias):
        delattr(self._connections, alias)

    def __getitem__(self, alias):
        if hasattr(self._connections, alias):
            return getattr(self._connections, alias)

        conn = self.__create_connection(settings.REST_SEARCH_CONNECTIONS[alias])
        setattr(self._connections, alias, conn)
        return conn

    def __create_connection(self, config):
        kwargs = {
            "host": config["HOST"],
            "port": config.get("PORT", 9200),
            "use_ssl": config.get("USE_SSL", False),
        }

        if (
            "AWS_ACCESS_KEY" in config
            and "AWS_SECRET_KEY" in config
            and "AWS_REGION" in config
        ):

            kwargs["connection_class"] = RequestsHttpConnection
            kwargs["http_auth"] = AWSRequestsAuth(
                aws_access_key=config["AWS_ACCESS_KEY"],
                aws_secret_access_key=config["AWS_SECRET_KEY"],
                aws_host=config["HOST"],
                aws_region=config["AWS_REGION"],
                aws_service="es",
            )

        return Elasticsearch(**kwargs)


def get_elasticsearch(indexer):
    return connections["default"]


def queue_add(updates):
    """
    Adds items to the updates queue, in the form:

    {
        'Book': [1, 2],
    }
    """
    if getattr(settings, "SEARCH_UPDATES_ENABLED", True):
        for doc_type, pks in updates.items():
            if doc_type in QUEUES:
                QUEUES[doc_type].update(set(pks))
            else:
                QUEUES[doc_type] = set(pks)


def queue_flush():
    """
    Triggers a celery task for the queued updates, if any.
    """
    global QUEUES
    if QUEUES:
        from rest_search.tasks import patch_index

        # convert sets to lists, otherwise they are not JSON-serializable
        args = {}
        for doc_type, pks in QUEUES.items():
            args[doc_type] = list(pks)
        patch_index.delay(args)
        QUEUES.clear()


connections = ConnectionHandler()



# -*- coding: utf-8 -*-

from django.utils.deprecation import MiddlewareMixin

from rest_search import queue_flush


class FlushUpdatesMiddleware(MiddlewareMixin):
    """
    Middleware that flushes ElasticSearch updates.
    """

    def process_response(self, request, response):
        queue_flush()
        return response



# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='djangorestsearch',
    version='0.8.6',
    url='https://github.com/wemap/django-rest-search',
    license='BSD',
    description='ElasticSearch integration for Django.',
    long_description=open('README.rst', 'r').read(),
    author='Jeremy Lainé',
    author_email='jeremy@getwemap.com',
    packages=['rest_search'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Programming Language :: Python :: 3 :: Only",
    ],
    install_requires=[
        'aws-requests-auth>=0.3.0',
        'elasticsearch>=6.0.0,<7.0.0',
    ],
    python_requires='>=3',
)



# -*- coding: utf-8 -*-

from django import forms


class SearchMixin(object):
    """
    Mixin to help build ElasticSearch queries.
    """

    def get_query(self):
        """
        Returns the query to be executed by ElasticSearch.
        """
        return self._score_query(self._build_query())

    def _build_query(self):
        query = {"bool": {}}

        for kind in ["filter", "must", "must_not", "should"]:
            clauses = getattr(self, "get_%s_clauses" % kind)()
            if clauses:
                query["bool"][kind] = clauses

        if query["bool"]:
            return query
        else:
            return {"match_all": {}}

    def get_filter_clauses(self):
        return []

    def get_must_clauses(self):
        return []

    def get_must_not_clauses(self):
        return []

    def get_should_clauses(self):
        return []

    def _score_query(self, query):
        return query


class SearchForm(SearchMixin, forms.Form):
    """
    Base form for building ElasticSearch queries.
    """

    def __init__(self, data, context={}):
        self.context = context
        super(SearchForm, self).__init__(data)
